/*
 * GStreamer
 * Copyright (C) 2005 Thomas Vander Stichele <thomas@apestaart.org>
 * Copyright (C) 2005 Ronald S. Bultje <rbultje@ronald.bitfreak.net>
 * Copyright (C) 2009 Dmitry Fink <<dmitry.fink@palm.com>>
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 * Alternatively, the contents of this file may be used under the
 * GNU Lesser General Public License Version 2.1 (the "LGPL"), in
 * which case the following provisions apply instead of the ones
 * mentioned above:
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <sys/mman.h>

#include <camera.h>
#include <cam_mmap.h>

#include <gst/gst.h>

#include "gstv4l2buffer.h"
#include "gstv4l2src.h"
#include "v4l2_cam.h"

GST_DEBUG_CATEGORY_EXTERN(v4l2src_debug);
#define GST_CAT_DEFAULT v4l2src_debug


static GstBufferClass *buffer_parent_class = NULL;

static void
gst_v4l2buffer_finalize(Gstv4l2buffer *buffer)
{
	int rc;
	Gstv4l2src* v4l2src = GST_V4L2SRC(buffer->owner);
	GST_DEBUG("gst_v4l2buffer_finalize fd=%d buffer=%p", buffer->fd, buffer->addr);

	// Reference the buffer before it is released completely
	gst_buffer_ref(GST_BUFFER(buffer));

	rc = v4l2_put_frame(v4l2src->fd, buffer->addr, buffer->cbcr_off, buffer->fd);
	
	// Call parent finalize
	GST_MINI_OBJECT_CLASS(buffer_parent_class)->finalize(GST_MINI_OBJECT(buffer));
}

static void
gst_v4l2buffer_init(Gstv4l2buffer *buffer, gpointer g_class)
{
}

static void
gst_v4l2buffer_class_init(gpointer g_class, gpointer class_data)
{
	GstMiniObjectClass *klass = GST_MINI_OBJECT_CLASS(g_class);

	buffer_parent_class = g_type_class_peek_parent(g_class);
	klass->finalize = (GstMiniObjectFinalizeFunction)gst_v4l2buffer_finalize;
}

static const GTypeInfo v4l2buffer_info = {
	sizeof(GstBufferClass),
	NULL,
	NULL,
	gst_v4l2buffer_class_init,
	NULL,
	NULL,
	sizeof(Gstv4l2buffer),
	0,
	(GInstanceInitFunc)gst_v4l2buffer_init,
	NULL
};

GType
gst_v4l2buffer_get_type(void)
{
	static GType type = 0;

	if (G_UNLIKELY(type == 0)) {
		type = g_type_register_static(GST_TYPE_BUFFER,
							       "Gstv4l2buffer",
							       &v4l2buffer_info,
							       0);
	}

	return type;
}
