/*
 * GStreamer
 * Copyright (C) 2005 Thomas Vander Stichele <thomas@apestaart.org>
 * Copyright (C) 2005 Ronald S. Bultje <rbultje@ronald.bitfreak.net>
 * Copyright (C) 2009 Dmitry Fink <<dmitry.fink@palm.com>>
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 *
 * Alternatively, the contents of this file may be used under the
 * GNU Lesser General Public License Version 2.1 (the "LGPL"), in
 * which case the following provisions apply instead of the ones
 * mentioned above:
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/**
 * SECTION:element-v4l2src
 *
 * FIXME:Describe v4l2src here.
 *
 * <refsect2>
 * <title>Example launch line</title>
 * |[
 * gst-launch -v -m fakesrc ! v4l2src ! fakesink silent=TRUE
 * ]|
 * </refsect2>
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <media/msm_camera_sensor.h>
#include <camera.h>
#include <cam_mmap.h>

#include <gst/gst.h>

#include <sys/time.h>



#include "gstv4l2src.h"
#include "v4l2_cam.h"

GST_DEBUG_CATEGORY(v4l2src_debug);
#define GST_CAT_DEFAULT v4l2src_debug

/* Filter signals and args */
enum
{
  /* FILL ME */
  LAST_SIGNAL
};

enum
{
	PROP_0,
	PROP_QUEUE_SIZE,
	PROP_TAKE_PICTURE,
	PROP_SET_PRV_DURATION,
	PROP_RESUME_PREVIEW,
	PROP_NUM,
	PROP_BURSTMODE,
	PROP_INTERVAL,
	PROP_FLASH,
	PROP_STATUS,
  	PROP_SILENT
};

typedef enum
{
	CAPTURE_SUCCESS       =  0,
	CAPTURE_PENDING       =  1,
	CAPTURE_ERR_FILE      =  2,	// error in file saving
	CAPTURE_ERR_DATA      =  3,	// gstreamer error and OMX erro
	CAPTURE_ERR_SYSTEM    =  4,	// V4l2 driver or system related e
	CAPTURE_START	      =	 5,
	ENABLE_FLASHON 	      =  6,
	DISABLE_FLASHON	      =  7
	
} app_errCode_type;

typedef enum
{
	CAM_STATE_IDLE,
	CAM_STATE_PREVIEW,
	CAM_STATE_PREFLASH,
	CAM_STATE_SNAPSHOT,
	CAM_STATE_TP
} cam_state_type;



enum{
 	FLASH_OFF,
  	FLASH_ON,
   	AUTO,
    	
};

/* the capabilities of the inputs and outputs.
 *
 * describe the real formats here.
 */

GST_BOILERPLATE (Gstv4l2src, gst_v4l2src, GstPushSrc, GST_TYPE_PUSH_SRC);

static void gst_v4l2src_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec);
static void gst_v4l2src_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec);

static gboolean gst_v4l2src_start(GstBaseSrc * src);
static gboolean gst_v4l2src_stop(GstBaseSrc * src);
static gboolean gst_v4l2src_set_caps(GstBaseSrc * src, GstCaps * caps);
static gboolean gst_v4l2src_send_appmsg(Gstv4l2src * v4l2src, app_errCode_type errCode);
//static GstCaps *gst_v4l2src_get_caps(GstBaseSrc * src);
static gboolean gst_v4l2src_query(GstBaseSrc * bsrc, GstQuery * query);
static GstFlowReturn gst_v4l2src_create(GstPushSrc * src, GstBuffer ** out);
static GstStateChangeReturn gst_v4l2src_change_state(GstElement* element, GstStateChange transition);

static unsigned clp2(unsigned x)
{
  x = x - 1;
  x = x | (x >> 1);
  x = x | (x >> 2);
  x = x | (x >> 4);
  x = x | (x >> 8);
  x = x | (x >>16);
  return x + 1;
}

/*===========================================================================
 * FUNCTION    - do_mmap -  
 *
 * DESCRIPTION:  retured virtual addresss
 *==========================================================================*/
byte *do_mmap (uint32 size, int *pmemFd)
{
	void     *ret; /* returned virtual address */
	int    pmem_fd   = 0;
	
	pmem_fd = open("/dev/pmem_adsp", O_RDWR);
	
	if (pmem_fd < 0) {
		GST_ERROR( "do_mmap: Open device /dev/pmem_adsp failed!");
		return NULL;
	}
	
	/* to make it page size aligned */
	size = (size + 4095) & (~4095);
	size = clp2(size);
	
	GST_DEBUG("do_mmap: pmem mmap size:%ld",size);
	
	ret = mmap(NULL,
				size,
				PROT_READ  | PROT_WRITE,
				MAP_SHARED,
				pmem_fd,
				0);
	
	if (ret == MAP_FAILED) {
		GST_ERROR("do_mmap: pmem mmap() failed: %s (%d)", strerror(errno), errno); 
		return NULL;
	}
	
	*pmemFd = pmem_fd;

	return (byte *)ret;
}

/*===========================================================================
 * FUNCTION    - do_munmap -  
 *
 * DESCRIPTION:
 *==========================================================================*/
int do_munmap (int pmem_fd, void *addr, size_t size)
{
	int rc;
	
	size = (size + 4095) & (~4095);
	size = clp2(size);
	
	GST_DEBUG("munmapped size = %d, virt_addr = 0x%x", size, (uint32)addr);
	
	rc = (munmap(addr, size));
	
	close(pmem_fd);
	
	return rc;
}

static int update_camstate(Gstv4l2src *v4l2src){

	int rc=0;
	
	pthread_mutex_lock(&v4l2src->skipframe_mutex);

	if(FLASH_OFF == v4l2src->flash_mode){			
		v4l2src->next_state = CAM_STATE_SNAPSHOT;
	}else if(FLASH_ON == v4l2src->flash_mode){				
		v4l2src->next_state = CAM_STATE_TP;
		
	}else if(AUTO == v4l2src->flash_mode){
		 rc = v4l2_readregister(0x301B);
		 GST_DEBUG("Luminace_Avg value 0x%x \n", rc);
		 if(rc < LUMINANCE_THRESHOLD){			
			v4l2src->next_state = CAM_STATE_TP ;
		 }else{				
			v4l2src->next_state = CAM_STATE_SNAPSHOT;	
		 }
		
	}else{

		GST_WARNING("Invalid CamState \n");

	}
	
	pthread_mutex_unlock(&v4l2src->skipframe_mutex);
	return 0;	
	
}


bool InitializelunaService(Gstv4l2src* v4l2src)
{
    bool retVal = false;
    LSError lserr;
    LSErrorInit(&lserr);

    retVal = LSRegister(NULL, &v4l2src->pServiceHandle, &lserr);
    if (retVal)
        retVal = LSGmainAttach(v4l2src->pServiceHandle, v4l2src->mainLoop, &lserr);

    return retVal;
}


bool DeinitializeLunaService(Gstv4l2src* v4l2src)

{

    bool retVal = false;

    LSError lserr;

    LSErrorInit(&lserr); 

    retVal = LSUnregister(v4l2src->pServiceHandle, &lserr); 

    return retVal;

}



//Call Backs

bool tpCallback_Start(LSHandle *sh, LSMessage *message, void *ctx){
	Gstv4l2src* v4l2src = (Gstv4l2src*)ctx;	
	struct json_object* root;
	bool result;
	
	
    	const char* str = LSMessageGetPayload(message);
    	if (str) {
	    GST_DEBUG("tpCallback_On Payload %s", str);	    
            root = json_tokener_parse(str);

            result = (root && !is_error(root));
	}

	if(result){
		struct json_object* label = json_object_object_get(root, "returnValue");
		if(label && !is_error(label) && json_object_get_boolean(label)){	
 		        GST_DEBUG("tp Started");
		    
		}
		
	}
	
	if(root && !is_error(root))
	   json_object_put(root);
	
	return true;
}



bool tpCallback_Off(LSHandle *sh, LSMessage *message, void *ctx){

	Gstv4l2src* v4l2src = (Gstv4l2src*)ctx;	
	struct json_object* root;
	bool result;
	
    	const char* str = LSMessageGetPayload(message);
    	if (str) {
	    GST_DEBUG("tpCallback_Off Payload %s", str);
            root = json_tokener_parse(str);

            result = (root && !is_error(root));
	}
	
	if(result){
		struct json_object* label = json_object_object_get(root, "returnValue");
		if(label && !is_error(label) && json_object_get_boolean(label)){
			pthread_mutex_lock(&v4l2src->mutex);
			v4l2src->next_state = CAM_STATE_PREFLASH;
			pthread_mutex_unlock(&v4l2src->mutex);
		}
	}

	if(root && !is_error(root))
	   json_object_put(root);
		
	return true;
}


bool tpCallback_On(LSHandle *sh, LSMessage *message, void *ctx){
	Gstv4l2src* v4l2src = (Gstv4l2src*)ctx;	
	struct json_object* root;
	bool result;
	
	
    	const char* str = LSMessageGetPayload(message);
    	if (str) {
	    GST_DEBUG("tpCallback_On Payload %s", str);	    
            root = json_tokener_parse(str);

            result = (root && !is_error(root));
	}

	if(result){
		struct json_object* label = json_object_object_get(root, "returnValue");
		if(label && !is_error(label) && json_object_get_boolean(label)){			
		       pthread_cond_broadcast(&v4l2src->tp_cond);
		    
		}
		
	}
	
	if(root && !is_error(root))
	   json_object_put(root);
	
	return true;
}


bool batteryCallback(LSHandle *sh, LSMessage *message, void *ctx){
	Gstv4l2src* v4l2src;
	v4l2src = (Gstv4l2src*)ctx;
	struct json_object* root;
	bool result;
	
    	const char* str = LSMessageGetPayload(message);
    	if (str) {
      	    GST_DEBUG("Battery Payload %s", str);
            root = json_tokener_parse(str);
            result = (root && !is_error(root));

    	}

   	if (result){
            struct json_object* label = json_object_object_get(root, "voltage_mV");

            if (label) {
		pthread_mutex_lock(&v4l2src->vl_mutex);
              	v4l2src->voltage_level  = json_object_get_int(label);		
		/*
		 GST_DEBUG("Voltage Level %d \n", v4l2src->voltage_level);
	     	 if(v4l2src->voltage_level < BATTERY_VOLTAGE_THRESHOLD){
			GST_DEBUG("voltage_level < BATTERY_VOLTAGE_THRESHOLD ");			
			v4l2src->voltage_safe = FALSE;
			gst_v4l2src_send_appmsg(v4l2src,DISABLE_FLASHON);
			
	      	}else if(!v4l2src->voltage_safe){
			GST_DEBUG("voltage_level > BATTERY_VOLTAGE_THRESHOLD ");
			v4l2src->voltage_safe = TRUE;
			gst_v4l2src_send_appmsg(v4l2src, ENABLE_FLASHON);
		}
		*/
		pthread_mutex_unlock(&v4l2src->vl_mutex);

 	    }
           

    	}
	
	if(root && !is_error(root))
	   json_object_put(root);
	return true;


}



static void
gst_v4l2src_base_init (gpointer gclass)
{
	GstElementClass *element_class = GST_ELEMENT_CLASS (gclass);
	GstCaps *caps_video;
	
	gst_element_class_set_details_simple(element_class,
		"v4l2src",
		"v4l2 element",
		"v4l2 video/still source",
		"Dmitry Fink <<dmitry.fink@palm.com>>");
	
	// Add template for video pad
	// Currently only one format is supported
	caps_video = gst_caps_new_simple ("video/x-raw-yuv",
		"format", GST_TYPE_FOURCC, GST_MAKE_FOURCC ('N', 'V', '1', '2'),
		"framerate", GST_TYPE_FRACTION, 30, 1,
		"width", G_TYPE_INT, VIDEO_WIDTH,
		"height", G_TYPE_INT, VIDEO_HEIGHT,
		"angle", G_TYPE_INT, 270,
		NULL);
	gst_element_class_add_pad_template
	    (element_class,
	     gst_pad_template_new("src", GST_PAD_SRC, GST_PAD_ALWAYS, caps_video));
	
	// Add template for the still pad
	caps_video = gst_caps_new_simple ("video/x-raw-yuv",
		"format", GST_TYPE_FOURCC, GST_MAKE_FOURCC ('N', 'V', '1', '2'),
		"width", G_TYPE_INT, SNAPSHOT_WIDTH,
		"height", G_TYPE_INT, SNAPSHOT_HEIGHT,
		"angle", G_TYPE_INT, 270,
		NULL);
	gst_element_class_add_pad_template
	    (element_class,
	     gst_pad_template_new("src_still", GST_PAD_SRC, GST_PAD_ALWAYS, caps_video));
}

/* initialize the v4l2src's class */
static void
gst_v4l2src_class_init (Gstv4l2srcClass * klass)
{
	GObjectClass *gobject_class;
	GstBaseSrcClass *basesrc_class;
	GstPushSrcClass *pushsrc_class;
	GstElementClass *element_class;
	
	gobject_class = (GObjectClass *) klass;
	basesrc_class = GST_BASE_SRC_CLASS(klass);
	pushsrc_class = GST_PUSH_SRC_CLASS(klass);
	element_class = (GstElementClass*) klass;
	
	gobject_class->set_property = gst_v4l2src_set_property;
	gobject_class->get_property = gst_v4l2src_get_property;
	
	g_object_class_install_property (gobject_class, PROP_SILENT,
		g_param_spec_boolean ("silent", "Silent", "Produce verbose output ?",
			FALSE, G_PARAM_READWRITE));
	g_object_class_install_property(gobject_class, PROP_TAKE_PICTURE,
					g_param_spec_boolean("take-picture",
									"Take Picture",
									"Take a still picture",
									FALSE,
									G_PARAM_WRITABLE));

	g_object_class_install_property(gobject_class, PROP_RESUME_PREVIEW,
					g_param_spec_boolean("resume-preview",
									"Resume Preview",
									"Resume Preview after still capture",
									FALSE,
									G_PARAM_WRITABLE));

	g_object_class_install_property(gobject_class, PROP_QUEUE_SIZE,
					g_param_spec_uint("queue-size",
									"Queue size",
									"Number of buffers to be enqueud in the driver",
									PREVIEW_FRAMES_NUM,
									PREVIEW_FRAMES_NUM,
									PREVIEW_FRAMES_NUM,
									G_PARAM_READWRITE));

	g_object_class_install_property(gobject_class, PROP_SET_PRV_DURATION,
					g_param_spec_uint("preview-duration",
									"Preview Duration",
									"Preview Duration of still captured frame in milli seconds",
									0, 0xffffffff, 0,
									G_PARAM_READWRITE));

	g_object_class_install_property(gobject_class, PROP_BURSTMODE,
					g_param_spec_boolean("burst-mode",
									"Burst Mode",
									"Enable burst mode for still capture",
									FALSE,
									G_PARAM_WRITABLE));

	g_object_class_install_property(gobject_class, PROP_NUM,
					g_param_spec_uint("num",
									"Burst length",
									"# of images are captured in a burst",
									0, 10, 2,
									G_PARAM_READWRITE));

	g_object_class_install_property(gobject_class, PROP_FLASH,
					g_param_spec_uint("flash-ctrl",
									"flash Control",
									"Command to control the flash",
									0, 2, 0,
									G_PARAM_READWRITE));

	g_object_class_install_property(gobject_class, PROP_STATUS,
					g_param_spec_uint("cam-stat",
									"camsrc Status ",
									"Status of the camera source",
									0, 3, 1,
									G_PARAM_READABLE));

	g_object_class_install_property(gobject_class, PROP_INTERVAL,
					g_param_spec_uint("interval",
									"Burst Interval",
									"Time Interval between bursts in msec.",
									0, 0xffffffff, 1000,
									G_PARAM_READWRITE));
	
//  basesrc_class->get_caps = GST_DEBUG_FUNCPTR(gst_v4l2src_get_caps);
//  basesrc_class->set_caps = GST_DEBUG_FUNCPTR(gst_v4l2src_set_caps);
	basesrc_class->start = GST_DEBUG_FUNCPTR(gst_v4l2src_start);
	basesrc_class->stop = GST_DEBUG_FUNCPTR(gst_v4l2src_stop);
//  basesrc_class->query = GST_DEBUG_FUNCPTR(gst_v4l2src_query);
//  basesrc_class->fixate = GST_DEBUG_FUNCPTR(gst_v4l2src_fixate);
	element_class->change_state = GST_DEBUG_FUNCPTR(gst_v4l2src_change_state);
	
	pushsrc_class->create = GST_DEBUG_FUNCPTR(gst_v4l2src_create);
}

/* initialize the new element
 * instantiate pads and add them to element
 * set pad calback functions
 * initialize instance structure
 */
static void
gst_v4l2src_init (Gstv4l2src * filter,
    Gstv4l2srcClass * gclass)
{
	// Add additional still pad
	filter->srcpad_still = gst_pad_new_from_template(
			gst_element_class_get_pad_template((GstElementClass *) gclass, 
			"src_still"), "src_still");
	gst_element_add_pad (GST_ELEMENT (filter), filter->srcpad_still);
	filter->silent = FALSE;

	gst_base_src_set_live(GST_BASE_SRC(filter), TRUE);

	filter->state = CAM_STATE_IDLE;
}

static gboolean gst_v4l2src_send_appmsg(Gstv4l2src *v4l2src, app_errCode_type errCode)
{
	char *desc;

	switch (errCode) {
		case CAPTURE_START: 
			desc = "CaptureStart"; 
			break;
		case CAPTURE_SUCCESS: 
			desc = "CaptureComplete"; 
			break;				
		case ENABLE_FLASHON:
			desc = "FlashOn";
			break;
		case DISABLE_FLASHON:
			desc = "FlashOff";
			break;
		default:
			g_assert(0);
	}

	GST_DEBUG("sending app msg %s", desc);
	GstMessage *msg = gst_message_new_application((GstObject *) v4l2src,
			gst_structure_new(desc, "Status", G_TYPE_INT, errCode, NULL));
	GstBus *bus = gst_element_get_bus((GstElement *) v4l2src);
	gst_bus_post(bus, msg);
	gst_object_unref(bus);

	return TRUE;
}

static GstFlowReturn
gst_v4l2src_take_picture(GstPushSrc * src)
{
	boolean rc;
	unsigned int addr;
    unsigned int cbcr_off;
    int fd;
    int ret;
    struct timeval ts;
	
	Gstv4l2src* v4l2src = GST_V4L2SRC(src);	
	GST_INFO_OBJECT(v4l2src, "gst_v4l2src_take_picture");

	gst_v4l2src_send_appmsg(v4l2src, CAPTURE_START);

	rc = v4l2_stop_preview(v4l2src->fd);
	
	if (rc){
				
		if(v4l2src->turn_flash){		   		   
		   if(v4l2src->voltage_level > BATTERY_VOLTAGE_THRESHOLD)  		
		   	rc = v4l2_cameraflashcontrol_ex(FLASH_CURRENT, MAINFLASH);	
		   else	
			rc = v4l2_cameraflashcontrol_ex(LOWVOLTAGE_FLASHCURRENT, MAINFLASH);
		}		
	}
	if (rc)
		rc = v4l2_start_snapshot(v4l2src->fd);

	if (rc)
		rc = v4l2_get_picture(v4l2src->fd);
	
	
	
	if(v4l2src->turn_flash)			
		v4l2_cameraflashcontrol (0);			
	


	if (rc)
		rc = v4l2_stop_snapshot(v4l2src->fd);

	
	ret = v4l2_get_frame(v4l2src->fd, &addr, &cbcr_off, &fd, &ts);
	if(!ret){
		GST_DEBUG("Previous frame is not returned \n");
		v4l2_put_frame(v4l2src->fd, addr, cbcr_off, fd);
	}
		
	
	if (rc){
		rc = v4l2_start_preview(v4l2src->fd);
		v4l2src->skip_framecount = SKIP_FRAMECOUNT;
	
	}
	
	//Assuming that flash is indeed turned off at this point
	if(v4l2src->turn_flash){
		pthread_mutex_lock(&v4l2src->tp_mutex);
		LSCall(v4l2src->pServiceHandle, URI_TP_SERVICE, TP_FLASH_OFF, 
			tpCallback_On, v4l2src, NULL, &v4l2src->lserror);
		pthread_cond_wait(&v4l2src->tp_cond, &v4l2src->tp_mutex);		
		pthread_mutex_unlock(&v4l2src->tp_mutex);	
		v4l2src->turn_flash = FALSE;
		GST_DEBUG("TP Enabled");
	}
	
	if (rc) {
		GST_INFO_OBJECT(v4l2src, "Snapshot taken with no errors");
		gst_v4l2src_send_appmsg(v4l2src, CAPTURE_SUCCESS);
	}
	else
		GST_WARNING_OBJECT(v4l2src, "Error taking snapshot");

	return (rc)?GST_FLOW_OK:GST_FLOW_ERROR;
}

static void
gst_v4l2src_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec)
{
	Gstv4l2src *v4l2src = GST_V4L2SRC (object);
	
	GST_DEBUG_OBJECT(object, "gst_v4l2src_set_property");
		
	switch (prop_id) {
		case PROP_QUEUE_SIZE:
			break;

		case PROP_TAKE_PICTURE:
			GST_DEBUG("PROP TAKE PICTURE \n");
			pthread_mutex_lock(&v4l2src->mutex);
			update_camstate(v4l2src);
			pthread_mutex_unlock(&v4l2src->mutex);
			break;

		case PROP_SET_PRV_DURATION:
			break;

		case PROP_RESUME_PREVIEW:
			GST_INFO_OBJECT(v4l2src, "PROP_RESUME_PREVIEW deprecated, ignoring");
			break;

		case PROP_BURSTMODE:
			break;

		case PROP_NUM:
			break;

		case PROP_INTERVAL:
			break;

		case PROP_FLASH:
			v4l2src->flash_mode = g_value_get_uint(value);
			break;

		case PROP_SILENT:
			v4l2src->silent = g_value_get_boolean (value);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		break;
	}
}

static void
gst_v4l2src_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec)
{
	Gstv4l2src *v4l2src = GST_V4L2SRC (object);
	
	GST_DEBUG_OBJECT(object, "gst_v4l2src_get_property");
	
	switch (prop_id) {
		case PROP_QUEUE_SIZE:
			g_value_set_uint(value, PREVIEW_FRAMES_NUM);
			break;
		case PROP_BURSTMODE:
			g_value_set_boolean(value, FALSE);
			break;
		case PROP_NUM:
			g_value_set_uint(value, 2);
			break;
		case PROP_INTERVAL:
			g_value_set_uint(value, 500);
			break;
		case PROP_FLASH:
			g_value_set_uint(value, 0);
			break;
		case PROP_STATUS:
			/* deprecated */
			g_value_set_uint(value, 0);
			break;
		case PROP_SET_PRV_DURATION:
			g_value_set_uint(value, 0);
			break;
		case PROP_SILENT:
			g_value_set_boolean (value, v4l2src->silent);
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		break;
	}
}

static gboolean gst_v4l2src_start(GstBaseSrc * src)
{
	int index;
	Gstv4l2src* v4l2src = GST_V4L2SRC(src);

	GST_DEBUG("gst_v4l2src_start");

	v4l2_interface_init(&v4l2src->fd);

	start_camconfig();

	v4l2src->dimension = malloc(sizeof(cam_ctrl_dimension_t));
	if (!v4l2src->dimension) {
		GST_ERROR_OBJECT(v4l2src, "main: malloc failed!");
		return FALSE;
	}

	memset(v4l2src->dimension, 0, sizeof(cam_ctrl_dimension_t)); 
	
	v4l2src->frame_count = 0;

	v4l2src->dimension->picture_width       = SNAPSHOT_WIDTH; 
	v4l2src->dimension->picture_height      = SNAPSHOT_HEIGHT;
	v4l2src->dimension->display_width       = VIDEO_WIDTH;
	v4l2src->dimension->display_height      = VIDEO_HEIGHT;
	v4l2src->dimension->ui_thumbnail_width  = THUMBNAIL_WIDTH;
	v4l2src->dimension->ui_thumbnail_height = THUMBNAIL_HEIGHT;
	
	v4l2src->preflash_frame_count = 0;
	v4l2src->turn_flash = FALSE;
	v4l2src->retry_count = 0;
	v4l2src->skip_framecount = 0;
	v4l2src->last_video_buffer_ts = 0;

	pthread_mutex_init (&v4l2src->mutex, NULL);

	pthread_mutex_init (&v4l2src->skipframe_mutex, NULL);
	

	
	v4l2src->mainLoop = g_main_loop_new(NULL, FALSE);
	InitializelunaService(v4l2src);

	pthread_mutex_init(&v4l2src->tp_mutex, NULL);
  	pthread_cond_init(&v4l2src->tp_cond, NULL);

	v4l2src->voltage_safe = TRUE;


	//Register for battery status
	LSCall(v4l2src->pServiceHandle, URI_SIGNAL_ADDMATCH, JSON_BATTERY_SIGNAL_ADDMATCH, 
	       batteryCallback, v4l2src, NULL, &v4l2src->lserror);

	LSSignalSend(v4l2src->pServiceHandle, URI_POWERD_BATTERY_SIGNAL_REQUEST, 
		     JSON_SIGNAL_REQUEST, &v4l2src->lserror);
	

	//Turn flash off and enable tp
	v4l2_cameraflashcontrol (0);
	
	LSCall(v4l2src->pServiceHandle, URI_TP_SERVICE, TP_FLASH_OFF, 
			tpCallback_Start, v4l2src, NULL, &v4l2src->lserror);
	
	if (v4l2_set_dimension(v4l2src->fd, v4l2src->dimension) == TRUE) {

		boolean activeBuffer;

		for (index = 0; index < PREVIEW_FRAMES_NUM; index++) {

			Gstv4l2buffer* buffer = (Gstv4l2buffer *)gst_mini_object_new(GST_TYPE_V4L2BUFFER);
			pmem_footer* footer;

			buffer->owner = (void*)v4l2src;

			buffer->addr = (unsigned long)do_mmap(VIDEO_BUFFER_SIZE + sizeof(pmem_footer), &(buffer->fd));
			buffer->cbcr_off = v4l2src->dimension->display_width * v4l2src->dimension->display_height;
		
			if (buffer->addr == 0) {
				GST_ERROR_OBJECT(v4l2src,"main: malloc failed!");
				goto ERROR;
			}

			if (index <= PREVIEW_PREBUFFER_FRAMES_NUM-1) {
				activeBuffer = TRUE; 
			} else {
				activeBuffer = FALSE; 
			}

			GST_DEBUG_OBJECT(v4l2src, "do_mmap pbuf = 0x%lx, pmem_fd = %d, active = %d", 
					buffer->addr, buffer->fd, activeBuffer);

			GST_BUFFER_SIZE(buffer) = VIDEO_BUFFER_SIZE;
			GST_BUFFER_MALLOCDATA(buffer) = NULL;
			GST_BUFFER_DATA(buffer) = buffer->addr;

			footer = (pmem_footer*)(buffer->addr + VIDEO_BUFFER_SIZE);
			footer->magic1 = PMEM_FOOTER_MAGIC1;
			footer->fd = buffer->fd;
			footer->addr = buffer->addr;
			footer->magic2 = PMEM_FOOTER_MAGIC2;

			v4l2_register_preview_bufs(v4l2src->fd,
								index,
								v4l2src->dimension, 
								buffer->addr,
								buffer->fd,
								activeBuffer);

			v4l2src->buffers[index] = buffer;
		}

		{
			// Allocate and set up still buffers (both main and thumbnail)
			GstSnapshotBuffer *buffer = (GstSnapshotBuffer *)gst_mini_object_new(GST_TYPE_SNAPSHOTBUFFER);
			v4l2src->snapshot_buffer = buffer;
			
			buffer->owner = (void*)v4l2src;
	
			buffer->thumb_img_addr = do_mmap(THUMBNAIL_BUFFER_SIZE, &buffer->thumb_img_fd);
			
			GST_DEBUG_OBJECT(v4l2src, "do_mmap thumbnail pbuf = 0x%x, pmem_fd = %d", 
					buffer->thumb_img_addr, buffer->thumb_img_fd);
			if (buffer->thumb_img_addr == NULL)
				goto ERROR;
			
			buffer->main_img_addr = do_mmap(SNAPSHOT_BUFFER_SIZE, &buffer->main_img_fd);
	
			GST_DEBUG_OBJECT(v4l2src, "do_mmap snapshot pbuf = 0x%x, pmem_fd = %d",
								buffer->main_img_addr, buffer->main_img_fd);
			if (buffer->main_img_addr == NULL)
				goto ERROR;
			
			v4l2_register_snapshot_bufs(v4l2src->fd, 
										v4l2src->dimension, 
										buffer->thumb_img_fd, 
										buffer->main_img_fd, 
										buffer->thumb_img_addr, 
										buffer->main_img_addr);
	
			GST_BUFFER_SIZE(v4l2src->snapshot_buffer) = SNAPSHOT_BUFFER_SIZE;
			GST_BUFFER_MALLOCDATA(v4l2src->snapshot_buffer) = NULL;
			GST_BUFFER_DATA(v4l2src->snapshot_buffer) = buffer->main_img_addr;
		}

	}

ERROR:
	return TRUE;
}

static gboolean gst_v4l2src_stop(GstBaseSrc * src)
{
	int rc, index;
	Gstv4l2src* v4l2src = GST_V4L2SRC(src);

	GST_DEBUG("gst_v4l2src_stop");

	// Unregister and free video/preview buffers
	for (index = 0; index < PREVIEW_FRAMES_NUM; ++index )
	{	
		Gstv4l2buffer* buffer = v4l2src->buffers[index];

		GST_DEBUG_OBJECT(v4l2src, "unregisterPreviewBuf %d", index);
		v4l2_unregister_preview_bufs(v4l2src->fd, 
									v4l2src->dimension, 
									buffer->fd, 
									(byte *)buffer->addr);
		GST_DEBUG_OBJECT(v4l2src, "do_munmap preview buffer %d, fd=%d, prev_buf=0x%x, size=%d", 
				index, buffer->fd, buffer->addr, VIDEO_BUFFER_SIZE + sizeof(pmem_footer));

		rc = do_munmap(buffer->fd, (void *)buffer->addr, VIDEO_BUFFER_SIZE + sizeof(pmem_footer));
		GST_DEBUG_OBJECT(v4l2src, "do_munmap done with return value %d", rc);
	}
	
	// Unregister and free snapshot buffers
	{
		GstSnapshotBuffer* buffer = v4l2src->snapshot_buffer;

		v4l2_unregister_snapshot_bufs(v4l2src->fd,
										v4l2src->dimension,
										buffer->thumb_img_fd, buffer->main_img_fd,  
										buffer->thumb_img_addr, buffer->main_img_addr);
		
		GST_DEBUG_OBJECT(v4l2src, "do_munmap thumbnail buffer fd=%d, prev_buf=0x%x, size=%d", 
				buffer->thumb_img_fd, buffer->thumb_img_addr, THUMBNAIL_BUFFER_SIZE);
		rc = do_munmap(buffer->thumb_img_fd, buffer->thumb_img_addr, THUMBNAIL_BUFFER_SIZE);
		GST_DEBUG_OBJECT(v4l2src, "do_munmap thumbnail buffer return value: %d", rc);

		GST_DEBUG_OBJECT(v4l2src, "do_munmap main image buffer fd=%d, prev_buf=0x%x, size=%d", 
				buffer->main_img_fd, buffer->main_img_addr, SNAPSHOT_BUFFER_SIZE);
		rc = do_munmap(buffer->main_img_fd, buffer->main_img_addr, SNAPSHOT_BUFFER_SIZE);
		GST_DEBUG_OBJECT(v4l2src, "do_munmap main image buffer return value: %d", rc);
	}

	if(v4l2src->dimension){
		free(v4l2src->dimension);
	}
	v4l2src->skip_framecount = 0;

	close(v4l2src->fd);

	stop_camconfig();

	v4l2_interface_destroy(&v4l2src->fd);

	DeinitializeLunaService(v4l2src);

	pthread_mutex_lock(&v4l2src->mutex);
	v4l2src->next_state = CAM_STATE_IDLE;	
	pthread_mutex_unlock(&v4l2src->mutex);

	
	return TRUE;
}

static GstFlowReturn gst_v4l2src_create(GstPushSrc * src, GstBuffer ** out)
{
	int rc;
	int index;
	boolean completed = FALSE;
	GstCaps *stillcapture_caps;
	unsigned int addr;
	unsigned int cbcr_off;
	int fd;
	struct timeval ts;
    GstClockTime gst_ts_current_frame;
    struct timeval res;

	Gstv4l2src* v4l2src = GST_V4L2SRC(src);

	GST_DEBUG("gst_v4l2src_create");
	GST_TIME_TO_TIMEVAL(GST_ELEMENT(v4l2src)->base_time, v4l2src->tv_start);

	while(!completed) {	
		v4l2src->current_state = v4l2src->next_state;
		switch(v4l2src->current_state){
			
			case CAM_STATE_IDLE:
				GST_DEBUG("CAM_STATE_IDLE");
				if(v4l2src->skip_framecount > 0)				
					pthread_mutex_unlock(&v4l2src->skipframe_mutex);

				if(v4l2src->turn_flash){
					
					v4l2_cameraflashcontrol (0);
					pthread_mutex_lock(&v4l2src->tp_mutex);
					LSCall(v4l2src->pServiceHandle, URI_TP_SERVICE, 	
					      TP_FLASH_OFF, tpCallback_On, v4l2src, NULL, &v4l2src->lserror);
					pthread_cond_wait(&v4l2src->tp_cond, 
						 	 &v4l2src->tp_mutex);		
					pthread_mutex_unlock(&v4l2src->tp_mutex);	
					v4l2src->turn_flash = FALSE;
					
				}				
				completed = TRUE;
				rc = GST_FLOW_WRONG_STATE;
				return rc;
				
			case CAM_STATE_PREFLASH:
				GST_DEBUG("PREFLASH STATE \n");			
				if( 1 ==  v4l2src->preflash_frame_count){				
					v4l2_cameraflashcontrol_ex(PREFLASH_CURRENT, PREFLASH);
					pthread_mutex_lock(&v4l2src->mutex);
					v4l2src->next_state = CAM_STATE_PREFLASH;
					pthread_mutex_unlock(&v4l2src->mutex);
	
				}else if(PREFLASH_FRAMECOUNT+1 == v4l2src->preflash_frame_count){
					v4l2src->preflash_frame_count = 0;				
					pthread_mutex_lock(&v4l2src->mutex);
					v4l2src->next_state = CAM_STATE_SNAPSHOT;
					pthread_mutex_unlock(&v4l2src->mutex);
										
				}

			case CAM_STATE_TP:
				
				if(CAM_STATE_TP == v4l2src->current_state){
					GST_DEBUG("CAM_STATE_TP- Calling luna TP service ");
					v4l2src->turn_flash = TRUE;		
					LSCall(v4l2src->pServiceHandle, URI_TP_SERVICE, TP_FLASH_ON, 
						tpCallback_Off, v4l2src, NULL, &v4l2src->lserror);		
					pthread_mutex_lock(&v4l2src->mutex);
					v4l2src->next_state = CAM_STATE_PREVIEW ;
					pthread_mutex_unlock(&v4l2src->mutex);


				}
	
			case  CAM_STATE_PREVIEW:
								
				rc = v4l2_get_frame(v4l2src->fd, &addr, &cbcr_off, &fd, &ts);
			
				if (rc < 0){

					if (rc == -ETIMEDOUT) {
						GST_DEBUG("gst_v4l2src_create: time-out receiving frame");
						
						// Check the internal state again, if it is not PREVIEW or PREFLASH
						// bailout with wrong state msg, otherwise report error
						if ((CAM_STATE_PREVIEW != v4l2src->current_state) &&
							(CAM_STATE_PREFLASH != v4l2src->current_state)) {
								return GST_FLOW_WRONG_STATE;
						}
						else {
							if(v4l2src->retry_count < MAX_RETRYCOUNT) {
								GST_DEBUG("gst_v4l2src_create: Retry count %d \n",            v4l2src->retry_count);
								v4l2src->retry_count++;
								rc = -EAGAIN;
								completed = FALSE;
													
							}else{						
								GST_ERROR("gst_v4l2src_create: Fatal error receving                frame \n");
								v4l2src->retry_count=0;
								rc = -EFAULT;
								completed = TRUE;			
								break;
							}
							
						}
					}
					
					// Got error, however may still be -EAGAIN
					if (rc != -EAGAIN) {
						// This is a real error, bail out!
						GST_ERROR("gst_v4l2src_create: fatal error receiving frame");
						completed = TRUE;
						break;
					}
				}
				else
				{	
					// Check the validity of the frame. Do we have one of these?
					// Find out the index inside our frame pool.
					//pthread_mutex_lock(&v4l2src->mutex_frame);
					for (index = 0; index < PREVIEW_FRAMES_NUM; index++)
					{
						if (v4l2src->buffers[index]->addr == addr) {
						
							timersub(&ts, &v4l2src->tv_start, &res);
							gst_ts_current_frame = GST_TIMEVAL_TO_TIME(res);
							
							GST_BUFFER_TIMESTAMP(v4l2src->buffers[index]) = gst_ts_current_frame;
							
							/* duration based on delta from previous frame, appears to be the most accurate */
							GST_BUFFER_DURATION(v4l2src->buffers[index]) = gst_ts_current_frame - v4l2src->last_video_buffer_ts;
							v4l2src->last_video_buffer_ts = gst_ts_current_frame;
							
							GST_BUFFER_OFFSET(v4l2src->buffers[index]) = v4l2src->frame_count;
							GST_BUFFER_OFFSET_END(v4l2src->buffers[index]) = v4l2src->frame_count+1;
							
							GST_DEBUG("setting timestamp %llu, on buffer index %d", GST_BUFFER_TIMESTAMP(v4l2src->buffers[index]), index);
							completed = TRUE;
							break;
						}
					}
					
					
					if (index >= PREVIEW_FRAMES_NUM)
					{
						// Got frame, but it doesn't match any frames in the pool
						GST_ERROR("gst_v4l2src_create error, got unknown frame from v4l2 fd=%d buffer=%p \n", fd, addr);
						rc = -1;
						completed = TRUE;
						break;	
					}
					if( CAM_STATE_PREFLASH == v4l2src->current_state ){				
						v4l2src->preflash_frame_count++;		
					}
					v4l2src->frame_count++;

					
					if (v4l2src->frame_count <= (PREVIEW_FRAMES_NUM - 
									PREVIEW_PREBUFFER_FRAMES_NUM ))
					
					{
						// Patch: Ideally we should register all available buffers to the qualcomm camera
						// and once kernel/dsp has a free buffer it should be sent to us, however
						// there seems to be some problem in the DSP and that is not whats' happenning,
						// first frame arrives, however despite the fact that DSP still has PREVIEW_FRAMES_NUM-1 buffers,
						// another one will be released only when we push the previous one back.
						// To overcome this problem, we initially register only PREVIEW_FRAMES_NUM-1 buffers, and once
						// first frame arrives, we push the last buffer as well. This solution seems to keep
						// buffers flowing...
						Gstv4l2buffer* buffer = v4l2src->buffers[v4l2src->frame_count+PREVIEW_PREBUFFER_FRAMES_NUM-1];
						v4l2_put_frame(v4l2src->fd, buffer->addr, buffer->cbcr_off, buffer->fd);
					}


					if(v4l2src->skip_framecount > 0){							
						v4l2_put_frame(v4l2src->fd, addr, cbcr_off, fd);
						GST_DEBUG("gst_v4l2src_create: Skipping Frame %d", 
							  (SKIP_FRAMECOUNT+1)-v4l2src->skip_framecount);		
						completed = FALSE ;
						v4l2src->skip_framecount--;
						if(v4l2src->skip_framecount <= 0)			
							pthread_mutex_unlock(&v4l2src->skipframe_mutex);						
										
						break;
					}
				
					v4l2src->retry_count=0;
					// Got valid frame!!!
					GST_DEBUG("gst_v4l2src_create index=%d, addr=%p", index, v4l2src->buffers[index]->addr);
					*out = v4l2src->buffers[index];
					completed = TRUE;
					
					
		
				}
				
				break;

			case   CAM_STATE_SNAPSHOT:
				GST_DEBUG("CAM_STATE_SNAPSHOT");
				GstFlowReturn ret = GST_FLOW_ERROR;		
				pthread_mutex_lock(&v4l2src->mutex);
				ret = gst_v4l2src_take_picture(v4l2src);
				if (GST_FLOW_OK == ret)
				{			
					GST_DEBUG("gst_v4l2src_create: setting still caps on the buffer");
					stillcapture_caps = gst_caps_new_simple ("video/x-raw-yuv",
							"format", GST_TYPE_FOURCC, GST_MAKE_FOURCC ('N', 'V', '1', '2'),
							"width", G_TYPE_INT, SNAPSHOT_WIDTH,
							"height", G_TYPE_INT, SNAPSHOT_HEIGHT,
							"angle", G_TYPE_INT, 270,
							NULL);
		
					gst_buffer_set_caps(v4l2src->snapshot_buffer, stillcapture_caps);
					ret = gst_pad_push(v4l2src->srcpad_still, v4l2src->snapshot_buffer);
					if (ret != GST_FLOW_OK) {
						GST_WARNING_OBJECT(v4l2src, "gst_pad_push() error: returned %d", ret);
						pthread_mutex_unlock(&v4l2src->mutex);
					}
					gst_caps_unref(stillcapture_caps);
				}
		
				// We go back to preview state				
				v4l2src->next_state = CAM_STATE_PREVIEW;
		
				completed = FALSE;				
				rc = 0;
				pthread_mutex_lock(&v4l2src->skipframe_mutex);				
				pthread_mutex_unlock(&v4l2src->mutex);		
				break ;

			default : 
				GST_DEBUG("DEFAULT STATE \n");
				completed = TRUE;
				rc = GST_FLOW_WRONG_STATE;
				return rc;
				break;
		
		}

		
	
	}
	
	return ((0 == rc)?GST_FLOW_OK:GST_FLOW_ERROR);

}

/* entry point to initialize the plug-in
 * initialize the plug-in itself
 * register the element factories and other features
 */
static gboolean
v4l2src_init (GstPlugin * v4l2src)
{
  /* debug category for fltering log messages
   *
   * exchange the string 'Template v4l2src' with your description
   */
  GST_DEBUG_CATEGORY_INIT (v4l2src_debug, "v4l2src",
      0, "Template v4l2src");

  return gst_element_register (v4l2src, "camsrc", GST_RANK_NONE,
      GST_TYPE_V4L2SRC);
}

/* gstreamer looks for this structure to register v4l2srcs
 *
 * exchange the string 'Template v4l2src' with your v4l2src description
 */
GST_PLUGIN_DEFINE (
    GST_VERSION_MAJOR,
    GST_VERSION_MINOR,
    "v4l2src",
    "Template v4l2src",
    v4l2src_init,
    VERSION,
    "unknown",
    "v4l2 plugin",
    "http://www.palm.com/"
)

static GstStateChangeReturn
gst_v4l2src_change_state(GstElement* element, GstStateChange transition)
{
	Gstv4l2src *v4l2src = GST_V4L2SRC(element);
	GstStateChangeReturn ret;

	GST_DEBUG("gst_v4l2src_change_state %s to %s",
			gst_element_state_get_name(GST_STATE_TRANSITION_CURRENT(transition)),
			gst_element_state_get_name(GST_STATE_TRANSITION_NEXT(transition)));

	switch (transition) {
		case GST_STATE_CHANGE_NULL_TO_READY:
			GST_DEBUG("GST_STATE_CHANGE_NULL_TO_READY");
			break;
		case GST_STATE_CHANGE_READY_TO_PAUSED:
			GST_DEBUG("GST_STATE_CHANGE_READY_TO_PAUSED");
			break;
		case GST_STATE_CHANGE_PAUSED_TO_PLAYING:
			GST_DEBUG("GST_STATE_CHANGE_PAUSED_TO_PLAYING");
			if (v4l2_start_preview(v4l2src->fd) == FALSE) {
				GST_ERROR_OBJECT(v4l2src, "gst_v4l2src_change_state: start_preview failed!");
				ret =  GST_STATE_CHANGE_FAILURE;
				goto exit;
			}
			pthread_mutex_lock(&v4l2src->mutex);
			v4l2src->next_state = CAM_STATE_PREVIEW;
			pthread_mutex_unlock(&v4l2src->mutex);			
			break;

		// Need to figure out why create is being called during transition in the first place
		// and even if it is valid, why can't we get another frame from the camera
		case GST_STATE_CHANGE_PLAYING_TO_PAUSED:
			GST_DEBUG("GST_STATE_CHANGE_PLAYING_TO_PAUSED");
			pthread_mutex_lock(&v4l2src->mutex);
			v4l2src->next_state = CAM_STATE_IDLE;
			pthread_mutex_unlock(&v4l2src->mutex);			
	  		break;

		default:
			break;
	}

	ret = GST_ELEMENT_CLASS(parent_class)->change_state(element, transition);
	if (ret != GST_STATE_CHANGE_SUCCESS && ret != GST_STATE_CHANGE_NO_PREROLL) {
		goto exit;
	}

	switch (transition) {
		case GST_STATE_CHANGE_PLAYING_TO_PAUSED:
			GST_DEBUG("GST_STATE_CHANGE_PLAYING_TO_PAUSED");
			if (v4l2_stop_preview(v4l2src->fd) == FALSE) {
				GST_ERROR_OBJECT(v4l2src, "gst_v4l2src_change_state: stop_preview failed!");
				ret =  GST_STATE_CHANGE_FAILURE;
				goto exit;
			}
			pthread_mutex_lock(&v4l2src->mutex);
			v4l2src->next_state = CAM_STATE_IDLE;
			pthread_mutex_unlock(&v4l2src->mutex);
			v4l2src->skip_framecount = 0;
	  		break;
		case GST_STATE_CHANGE_PAUSED_TO_READY:
			GST_DEBUG("GST_STATE_CHANGE_PAUSED_TO_READY");
			break;
		case GST_STATE_CHANGE_READY_TO_NULL:
			GST_DEBUG("GST_STATE_CHANGE_READY_TO_NULL");
			break;
		default:
			break;
	}

exit:
	return ret;
}
